From mathcomp Require Import all_ssreflect.

Require Import Psatz.

Set Implicit Arguments.
Unset Strict Implicit.
Unset Printing Implicit Defensive.

Declare ML Module "paramcoq".

Record cc_fun (T : Type) := {
  ccw : T -> T -> T -> bool;
}.

Record map_funs (T : Type) := {
  sort : Type;
  map0 : sort;
  eval : sort -> T -> T;
  update : sort -> T -> T -> sort;
  map_size : sort -> nat
}.

Section seq_algorithm.

Variables (P : Type) (p0 : P) (ccf : cc_fun P)
  (mf : map_funs P).

Notation ccw := (ccw ccf).
Notation update := update.
Notation map0 := (map0 mf).

Definition seq3 (s : seq P) :=
   let a := nth p0 s 0 in
   let b := nth a s 1 in
   let c := nth b s 2 in
   if ccw a b c then [:: a; b; c] else [:: b; a; c].


Definition map_cycle3 (s : seq P) :=
  match s with
  | [:: a; b; c] => update (update (update map0 c a) b c) a b
  | _ => map0
  end.

Definition inside_triangle_seq (trs : seq P) (p : P) :=
  match trs with
  | [:: a; b; c] => ccw a b p && ccw b c p && ccw c a p
  | _ => false
  end.

Fixpoint pick_triangle_seq (trs : seq (seq P)) (p : P) :=
  match trs with
  | tr :: trs =>
    if inside_triangle_seq tr p then Some(tr, trs)
    else
       match pick_triangle_seq trs p with
       | Some (v, others) => Some (v, tr::others)
       | None => None
       end
  | _ => None
  end.

Definition inside_triangles tr (p : P):=
  match tr with
  | [:: a; b; c] => [:: [:: p; a; b]; [:: p; b; c]; [:: p; c; a]]
  | _ => [::]
  end.

Fixpoint find_next_purple_map (p : P) (map : sort mf) side n b :=
  match n with
    0 => b
  | S k => if eqb (ccw b (eval map b) p) side then
             b
           else find_next_purple_map p map side k (eval map b)
  end.

Definition find_purple_points_map p (m : sort mf) (e : P) : P * P :=
  let n := map_size m in
  if ccw e (eval m e) p then
   let pp1 := find_next_purple_map p m false n (eval m e) in
     (pp1, find_next_purple_map p m true n (eval m pp1))
  else
   let pp2 := find_next_purple_map p m true n (eval m e) in
     (find_next_purple_map p m false n (eval m pp2), pp2).

Fixpoint add_red_triangles p (l : seq P) tr :=
  match l with
  | a :: (b :: _) as l' => 
    if ccw b a p then
      add_red_triangles p l' ([:: b; a; p] :: tr)
    else
      tr
  | _ => tr
  end.

Fixpoint map_orbit
  (eq_f : P -> P -> bool)
  (m : sort mf) (e stopper : P) (fuel : nat) trailer :
  seq P :=
  match fuel with
  | O => trailer
  | S k =>
    if eq_f (eval m e) stopper then
      e::trailer
    else
       e::map_orbit eq_f m (eval m e) stopper k trailer
  end.

Definition new_map (m : sort mf) (a b p : P) :=
  update (update m p b) a p.

Fixpoint naive5 eq_f (s : seq P) :
   seq (seq P) * (sort mf) * P :=
   match  s with
    nil =>  (nil, update map0 p0 p0, p0)
  | p :: tl => 
    if (size s <= 3)%nat then
      let tr1 := seq3 s in
      ([:: tr1], map_cycle3 tr1, head p0 tr1)
    else 
      let '(tr, chf, bp) := naive5 eq_f tl in
      if pick_triangle_seq tr p is Some (t, trmt) then
       (inside_triangles t p ++ trmt, chf, bp)
      else
        let (a, b) := find_purple_points_map p chf bp in
          (add_red_triangles p
            (map_orbit eq_f chf a a (map_size chf) [:: a]) tr,
           new_map chf a b p, p)
  end.

End seq_algorithm.

Parametricity seq.
Parametricity prod.
Parametricity bool.
Parametricity cc_fun.
Parametricity nat.
Parametricity map_funs.
Parametricity eq.
Parametricity False.
Parametricity reflect.
Parametricity Equality.mixin_of.
Parametricity eqType.
Parametricity map_orbit.
Parametricity new_map.
Parametricity map_cycle3.
Parametricity head.
Parametricity option.
Parametricity pick_triangle_seq.
Parametricity inside_triangles.
Parametricity True.
Parametricity find_next_purple_map.
Parametricity find_purple_points_map.
Parametricity add_red_triangles.
Parametricity map_size.
Parametricity naive5.
