From mathcomp Require Import all_ssreflect ssralg ssrnum ssrint.

From triangles Require Import triangles3 to_ccw_system.

Open Scope ring_scope.

Section fix_fin_type.

Variable R : realDomainType.

Variable l : seq (R * R).

Let P := seq_sub l.

Definition no_det_ccf : cc_fun [finType of P] :=
  @Build_cc_fun _
  (fun a' b' c' : P =>
    let a := val a' in let b := val b' in let c := val c' in
    let dt :=
        b.1 * c.2 - b.2 * c.1 - (a.1 * c.2 - a.2 * c.1) + a.1 * b.2
           - a.2 * b.1 in
     if 0 < dt then true
     else if 0 == dt then
       if a.1 < b.1 then (b.1 < c.1) || (c.1 < a.1)
       else if b.1 < a.1 then (b.1 < c.1) && (c.1 < a.1)
       else if a.2 < b.2 then (b.2 < c.2) || (c.2 < a.2)
       else if b.2 < a.2 then (b.2 < c.2) && (c.2 < a.2)
       else false
    else false).

Lemma no_det_ccfP (a b c : P) :
  triangles3.ccw no_det_ccf a b c = ccw (val a) (val b) (val c).
Proof.
by rewrite /ccw /= /surface Knuth_axiom5.expand_det_surface.
Qed.

End fix_fin_type.

Section assoc_map.

Variable P : eqType.

Variable default : P.

Definition lsort := list (P * P).

Definition lmap0 : lsort := nil.

Definition map_size : lsort -> nat := 
  fun m => (2 * (size m) + 3)%N.

Definition update : lsort -> P -> P -> lsort :=
  fun l k v =>
    (fix rec l :=
      match l with 
      | nil => (k, v) :: nil
      | (k', v')::tl => if k == k' then (k', v)::tl else (k', v')::rec tl
      end) l.

Definition eval : lsort -> P -> P :=
  fun l k =>
    (fix rec l :=
        match l with
        | nil => default
        | (k', v')::tl => if k == k' then v' else rec tl
        end) l.

Lemma leval_update1 : forall (m : lsort) x y, eval (update m x y) x = y.
Proof.
elim=> [ | [k v] tl IH] x y /=; case: ifP=> //.
    by rewrite eqxx.
  by move=> /eqP -> /=; rewrite eqxx.
by rewrite /= => ->.
Qed.

Lemma leval_update_dif : forall (m : lsort) x x' y, x != x' ->
    eval (update m x y) x' = eval m x'.
Proof.
elim=> [ | [k v] tl IH] x x' y /=; case: ifP=> //.
    by rewrite eq_sym => ->.
  by move=> /eqP -> /=; rewrite eq_sym => /negbTE ->.
by rewrite /= => xnk xn'; case:ifP=> _ //; apply: IH.
Qed.

Lemma map_n_default_pair (m : lsort) (x : P) :
  eval m x != default -> (x, (eval m x)) \in m.
Proof.
elim: m=> [|[x' v'] tl IH]; first by rewrite eqxx.
rewrite /=; case: ifP; rewrite !inE; first by move=> /eqP ->; rewrite eqxx.
by move=> xnx' vndef; rewrite IH ?orbT.
Qed.

Lemma lmap_sizeP (m : lsort) (x : P) (p : seq P):
                (fpath (eval m)) x p ->
                uniq (x :: p) ->
                (size (x :: p) <= map_size m)%N .
Proof.
move: p x.
have main p x : default \notin (x :: p) -> fpath (eval m) x p ->
   uniq (x :: p) -> (size (x :: p) <= (size m).+1)%N.
  move=> dnp pp up /=; rewrite ltnS.
  suff h : {subset p <= [seq i.2 | i <- m]}.
    have -> : size m = size [seq i.2 | i <- m] by rewrite size_map.
    by apply: uniq_leq_size => //; move: up; rewrite cons_uniq=> /andP[].
  move/pathP: pp => /= pp z /(nthP x) [i /= ilts <-] {z}.
  rewrite -(eqP (pp x i ilts)).
  have /map_n_default_pair it : eval m (nth x (x :: p) i) != default.
    rewrite (eqP (pp _ _ ilts)); apply/eqP=> abs; case/negP: dnp; rewrite -abs.
    by rewrite inE mem_nth ?orbT.
  by apply/mapP; exists (nth x (x :: p) i, eval m (nth x (x :: p) i)).
move=> p x; have [] := boolP (default \in x :: p); last first.
  move=> dnin pp up; move: (main _ _ dnin pp up) => pl.
  apply: (leq_trans pl); rewrite /map_size addnS ltnS mulSn mul1n.
  by rewrite -addnA leq_addr.
move=> din pp up; have/(nth_index x):= din; set i := index _ _.
have /= : (i < size (x :: p))%N by rewrite /i index_mem.
rewrite ltnS.
case: i => [ | i] /= i_s.
  move=> xdef; move: pp up; rewrite xdef.
  case: p din i_s => [ | y p] din _; first by rewrite /map_size addnS.
  move=> /= /andP [ed pp] /andP [dn up].
  have h : (size (y :: p) <= (size m).+1)%N by apply: main.
  rewrite /= /map_size mulSn mul1n addnS ltnS (leq_trans h) //.
  by rewrite addnS ltnS -!addnA leq_addr.
move=> idef.
have := cat_take_drop i.+1 (x :: p).
set p1 := take _ _; set p2 := drop _ _ => xpq.
have p2shape : p2 = default :: drop i.+1 p by rewrite -idef -drop_nth.
move: (xpq); rewrite /p1 /= => [[]] pq.
move: up; rewrite -xpq cat_uniq=> /andP[up1 /andP[nhas up2]].
have sp1 : (size p1 <= (size m).+1)%N.
rewrite /p1 -[take _ _]/(x :: take i p); apply: main=> //.
    by move: nhas; rewrite p2shape /= negb_or /p1 /==> /andP [].
  by move: (pp); rewrite -{1}pq cat_path => /andP[].
have sp2 : (size p2 <= (size m).+2)%N.
  rewrite p2shape; case cd : (drop i.+1 p) => [// | z p'].
  rewrite -[size [:: _, _ & _]]/(size (z :: p')).+1 ltnS.
  have uzp : uniq (z :: p').
    by move: up2; rewrite p2shape cons_uniq cd=> /andP[].
  apply: main => //.
    by move: up2; rewrite p2shape cons_uniq cd => /andP[].
  move: pp; rewrite -pq cat_path p2shape cd /=.
  by move=> /andP[] ? /andP[] ? /andP[] ?.
rewrite -pq size_cat -addSn -[S _]/(size p1) /map_size mulSn mul1n.
rewrite !addnS addn0 -addSn -2!addnS.
by apply: leq_add.    
Qed.

Definition assoc_list_map_funs : map_funs P :=
  @Build_map_funs P lsort nil eval update map_size.

Definition assoc_list_map_system : map_system assoc_list_map_funs.
constructor.
  exact leval_update1.
  exact leval_update_dif.
  exact lmap_sizeP.
Qed.

End assoc_map.

Lemma inl0 (T : eqType) (x : T) (l : seq T) :
  x \in x :: l.
Proof.
by rewrite inE eqxx.
Qed.

Definition map_to_orbit (T : eqType) (x : T) (mf : map_funs T) (m : sort mf):
  seq T :=
  x ::(fix aux (n : nat) (y stop : T) :=
   match n with
     0 => [::]
   | S p => if y == stop then [::] else y::aux p (triangles3.eval m y) stop end)
     (triangles3.map_size m) (triangles3.eval m x) x.

Definition int_naive5 :
 seq (int * int) -> (seq (seq (int * int)) * seq (int * int) * (int * int)):=
 fun l =>
   let l' := (0:int, 0:int) :: l in
   let p0 := (SeqSub (inl0 _ (0:int, 0:int) l)) in
   let lp := [seq (insubd p0 x) | x <- l] : seq (seq_sub l') in
   let ccf := no_det_ccf _ l' in
   let mf := assoc_list_map_funs [eqType of (seq_sub l')] p0 in
   let result :=
     naive5 p0 ccf mf lp in
     ([seq [seq val x | x <- t] | t <- result.1.1], 
      [seq val v | v <- map_to_orbit _ result.2 mf result.1.2],
      (val result.2)).

